@@ -28,7 +28,7 @@ Contributors:
 #include "logging_mosq.h"
 #include "memory_mosq.h"
 #include "messages_mosq.h"
-#include "mqtt3_protocol.h"
+#include "mqtt_protocol.h"
 #include "net_mosq.h"
 #include "packet_mosq.h"
 #include "read_handle.h"
@@ -38,24 +38,46 @@ Contributors:
 
 int handle__pubrel(struct mosquitto_db *db, struct mosquitto *mosq)
 {
+	uint8_t reason_code;
 	uint16_t mid;
 #ifndef WITH_BROKER
 	struct mosquitto_message_all *message = NULL;
 #endif
 	int rc;
+	mosquitto_property *properties = NULL;
 
 	assert(mosq);
-	if(mosq->protocol == mosq_p_mqtt311){
+
+	if(mosq->state != mosq_cs_connected){
+		return MOSQ_ERR_PROTOCOL;
+	}
+
+	if(mosq->protocol != mosq_p_mqtt31){
 		if((mosq->in_packet.command&0x0F) != 0x02){
 			return MOSQ_ERR_PROTOCOL;
 		}
 	}
 	rc = packet__read_uint16(&mosq->in_packet, &mid);
 	if(rc) return rc;
+	if(mid == 0) return MOSQ_ERR_PROTOCOL;
+
+	if(mosq->protocol == mosq_p_mqtt5 && mosq->in_packet.remaining_length > 2){
+		rc = packet__read_byte(&mosq->in_packet, &reason_code);
+		if(rc) return rc;
+
+		if(mosq->in_packet.remaining_length > 3){
+			rc = property__read_all(CMD_PUBREL, &mosq->in_packet, &properties);
+			if(rc) return rc;
+		}
+	}
+
 #ifdef WITH_BROKER
 	log__printf(NULL, MOSQ_LOG_DEBUG, "Received PUBREL from %s (Mid: %d)", mosq->id, mid);
 
-	rc = db__message_release(db, mosq, mid, mosq_md_in);
+	/* Immediately free, we don't do anything with Reason String or User Property at the moment */
+	mosquitto_property_free_all(&properties);
+
+	rc = db__message_release_incoming(db, mosq, mid);
 	if(rc == MOSQ_ERR_PROTOCOL){
 		return rc;
 	}else if(rc != MOSQ_ERR_SUCCESS){
@@ -63,9 +85,20 @@ int handle__pubrel(struct mosquitto_db *db, struct mosquitto *mosq)
 		 * due to a repeated PUBREL after a client has reconnected. */
 		log__printf(mosq, MOSQ_LOG_WARNING, "Warning: Received PUBREL from %s for an unknown packet identifier %d.", mosq->id, mid);
 	}
+
+	rc = send__pubcomp(mosq, mid);
+	if(rc) return rc;
 #else
+	UNUSED(db);
+
 	log__printf(mosq, MOSQ_LOG_DEBUG, "Client %s received PUBREL (Mid: %d)", mosq->id, mid);
 
+	rc = send__pubcomp(mosq, mid);
+	if(rc){
+		message__remove(mosq, mid, mosq_md_in, &message, 2);
+		return rc;
+	}
+
 	rc = message__remove(mosq, mid, mosq_md_in, &message, 2);
 	if(rc){
 		return rc;
@@ -78,12 +111,16 @@ int handle__pubrel(struct mosquitto_db *db, struct mosquitto *mosq)
 			mosq->on_message(mosq, mosq->userdata, &message->msg);
 			mosq->in_callback = false;
 		}
+		if(mosq->on_message_v5){
+			mosq->in_callback = true;
+			mosq->on_message_v5(mosq, mosq->userdata, &message->msg, properties);
+			mosq->in_callback = false;
+		}
 		pthread_mutex_unlock(&mosq->callback_mutex);
+		mosquitto_property_free_all(&properties);
 		message__cleanup(&message);
 	}
 #endif
-	rc = send__pubcomp(mosq, mid);
-	if(rc) return rc;
 
 	return MOSQ_ERR_SUCCESS;
 }
