@@ -92,7 +92,7 @@ static void temp__expire_websockets_clients(struct mosquitto_db *db)
 						}
 					}
 					/* Client has exceeded keepalive*1.5 */
-					do_disconnect(db, context);
+					do_disconnect(db, context, MOSQ_ERR_KEEPALIVE);
 				}
 			}
 		}
@@ -101,7 +101,7 @@ static void temp__expire_websockets_clients(struct mosquitto_db *db)
 }
 #endif
 
-int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int listensock_count, int listener_max)
+int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int listensock_count)
 {
 #ifdef WITH_SYS_TREE
 	time_t start_time = mosquitto_time();
@@ -110,7 +110,6 @@ int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int li
 	time_t last_backup = mosquitto_time();
 #endif
 	time_t now = 0;
-	time_t now_time;
 	int time_count;
 	int fdcount;
 	struct mosquitto *context, *ctxt_tmp;
@@ -216,8 +215,6 @@ int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int li
 		}
 #endif
 
-		now_time = time(NULL);
-
 		time_count = 0;
 		HASH_ITER(hh_sock, db->contexts_by_sock, context, ctxt_tmp){
 			if(time_count > 0){
@@ -238,7 +235,7 @@ int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int li
 							&& now > context->bridge->primary_retry){
 
 						if(context->bridge->primary_retry_sock == INVALID_SOCKET){
-							rc = net__try_connect(context, context->bridge->addresses[0].address,
+							rc = net__try_connect(context->bridge->addresses[0].address,
 									context->bridge->addresses[0].port,
 									&context->bridge->primary_retry_sock, NULL, false);
 
@@ -317,19 +314,11 @@ int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int li
 						pollfd_index++;
 #endif
 					}else{
-						do_disconnect(db, context);
+						do_disconnect(db, context, MOSQ_ERR_CONN_LOST);
 					}
 				}else{
-					if(db->config->connection_messages == true){
-						if(context->id){
-							id = context->id;
-						}else{
-							id = "<unknown>";
-						}
-						log__printf(NULL, MOSQ_LOG_NOTICE, "Client %s has exceeded timeout, disconnecting.", id);
-					}
 					/* Client has exceeded keepalive*1.5 */
-					do_disconnect(db, context);
+					do_disconnect(db, context, MOSQ_ERR_KEEPALIVE);
 				}
 			}
 		}
@@ -469,16 +458,16 @@ int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int li
 			}
 		}
 #endif
-		now_time = time(NULL);
-		if(db->config->persistent_client_expiration > 0 && now_time > expiration_check_time){
+		now = time(NULL);
+		if(db->config->persistent_client_expiration > 0 && now > expiration_check_time){
 			HASH_ITER(hh_id, db->contexts_by_id, context, ctxt_tmp){
-				if(context->sock == INVALID_SOCKET && context->clean_session == 0){
+				if(context->sock == INVALID_SOCKET && context->session_expiry_interval > 0 && context->session_expiry_interval != UINT32_MAX){
 					/* This is a persistent client, check to see if the
 					 * last time it connected was longer than
 					 * persistent_client_expiration seconds ago. If so,
 					 * expire it and clean up.
 					 */
-					if(now_time > context->disconnect_t+db->config->persistent_client_expiration){
+					if(now > context->session_expiry_time){
 						if(context->id){
 							id = context->id;
 						}else{
@@ -486,9 +475,9 @@ int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int li
 						}
 						log__printf(NULL, MOSQ_LOG_NOTICE, "Expiring persistent client %s due to timeout.", id);
 						G_CLIENTS_EXPIRED_INC();
-						context->clean_session = true;
-						context->state = mosq_cs_expiring;
-						do_disconnect(db, context);
+						context->session_expiry_interval = 0;
+						context__set_state(context, mosq_cs_expiring);
+						do_disconnect(db, context, MOSQ_ERR_SUCCESS);
 					}
 				}
 			}
@@ -555,6 +544,9 @@ int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int li
 			}
 		}
 #endif
+		now = time(NULL);
+		session_expiry__check(db, now);
+		will_delay__check(db, now);
 #ifdef WITH_PERSISTENCE
 		if(db->config->persistence && db->config->autosave_interval){
 			if(db->config->autosave_on_changes){
@@ -616,7 +608,7 @@ int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int li
 	return MOSQ_ERR_SUCCESS;
 }
 
-void do_disconnect(struct mosquitto_db *db, struct mosquitto *context)
+void do_disconnect(struct mosquitto_db *db, struct mosquitto *context, int reason)
 {
 	char *id;
 #ifdef WITH_EPOLL
@@ -628,8 +620,8 @@ void do_disconnect(struct mosquitto_db *db, struct mosquitto *context)
 	}
 #ifdef WITH_WEBSOCKETS
 	if(context->wsi){
-		if(context->state != mosq_cs_disconnecting){
-			context->state = mosq_cs_disconnect_ws;
+		if(context->state != mosq_cs_disconnecting && context->state != mosq_cs_disconnect_with_will){
+			context__set_state(context, mosq_cs_disconnect_ws);
 		}
 		if(context->wsi){
 			libwebsocket_callback_on_writable(context->ws_context, context->wsi);
@@ -654,8 +646,26 @@ void do_disconnect(struct mosquitto_db *db, struct mosquitto *context)
 			}else{
 				id = "<unknown>";
 			}
-			if(context->state != mosq_cs_disconnecting){
-				log__printf(NULL, MOSQ_LOG_NOTICE, "Socket error on client %s, disconnecting.", id);
+			if(context->state != mosq_cs_disconnecting && context->state != mosq_cs_disconnect_with_will){
+				switch(reason){
+					case MOSQ_ERR_SUCCESS:
+						break;
+					case MOSQ_ERR_PROTOCOL:
+						log__printf(NULL, MOSQ_LOG_NOTICE, "Client %s disconnected due to protocol error.", id);
+						break;
+					case MOSQ_ERR_CONN_LOST:
+						log__printf(NULL, MOSQ_LOG_NOTICE, "Socket error on client %s, disconnecting.", id);
+						break;
+					case MOSQ_ERR_AUTH:
+						log__printf(NULL, MOSQ_LOG_NOTICE, "Client %s disconnected, no longer authorised.", id);
+						break;
+					case MOSQ_ERR_KEEPALIVE:
+						log__printf(NULL, MOSQ_LOG_NOTICE, "Client %s has exceeded timeout, disconnecting.", id);
+						break;
+					default:
+						log__printf(NULL, MOSQ_LOG_NOTICE, "Socket error on client %s, disconnecting.", id);
+						break;
+				}
 			}else{
 				log__printf(NULL, MOSQ_LOG_NOTICE, "Client %s disconnected.", id);
 			}
@@ -668,19 +678,6 @@ void do_disconnect(struct mosquitto_db *db, struct mosquitto *context)
 		}
 #endif		
 		context__disconnect(db, context);
-#ifdef WITH_BRIDGE
-		if(context->clean_session && !context->bridge){
-#else
-		if(context->clean_session){
-#endif
-			context__add_to_disused(db, context);
-			if(context->id){
-				context__remove_from_by_id(db, context);
-				mosquitto__free(context->id);
-				context->id = NULL;
-			}
-		}
-		context->state = mosq_cs_disconnected;
 	}
 }
 
@@ -755,7 +752,7 @@ static void loop_handle_reads_writes(struct mosquitto_db *db, struct pollfd *pol
 				len = sizeof(int);
 				if(!getsockopt(context->sock, SOL_SOCKET, SO_ERROR, (char *)&err, &len)){
 					if(err == 0){
-						context->state = mosq_cs_new;
+						context__set_state(context, mosq_cs_new);
 #if defined(WITH_ADNS) && defined(WITH_BRIDGE)
 						if(context->bridge){
 							bridge__connect_step3(db, context);
@@ -764,12 +761,12 @@ static void loop_handle_reads_writes(struct mosquitto_db *db, struct pollfd *pol
 #endif
 					}
 				}else{
-					do_disconnect(db, context);
+					do_disconnect(db, context, MOSQ_ERR_CONN_LOST);
 					continue;
 				}
 			}
 			if(packet__write(context)){
-				do_disconnect(db, context);
+				do_disconnect(db, context, MOSQ_ERR_CONN_LOST);
 				continue;
 			}
 		}
@@ -811,7 +808,7 @@ static void loop_handle_reads_writes(struct mosquitto_db *db, struct pollfd *pol
 #endif
 			do{
 				if(packet__read(db, context)){
-					do_disconnect(db, context);
+					do_disconnect(db, context, MOSQ_ERR_CONN_LOST);
 					continue;
 				}
 			}while(SSL_DATA_PENDING(context));
@@ -821,7 +818,7 @@ static void loop_handle_reads_writes(struct mosquitto_db *db, struct pollfd *pol
 #else
 			if(context->pollfd_index >= 0 && pollfds[context->pollfd_index].revents & (POLLERR | POLLNVAL | POLLHUP)){
 #endif
-				do_disconnect(db, context);
+				do_disconnect(db, context, MOSQ_ERR_CONN_LOST);
 				continue;
 			}
 		}
